"""End-to-end tests for the command-line interface main functions, which will run most part of the codebase"""

import csv
import random
from zipfile import ZipFile

from outgram.cli import archive, collect_profile, collect_profile_posts, instagram_post
from outgram.instagram import Instagram

MIN_WAIT = 5.0
MAX_WAIT = 10.0
INSTAGRAM_USERNAME_ID = {
    "crio.cafe": "21057648761",
    "pythonbrasil": "1543793548",
    "estadao": "23456103",
}
INSTAGRAM_USERNAMES = list(INSTAGRAM_USERNAME_ID.keys())
INSTAGRAM_USER_IDS = list(INSTAGRAM_USERNAME_ID.values())
INSTAGRAM_USER_MIX = []
for index, (username, user_id) in enumerate(INSTAGRAM_USERNAME_ID.items()):
    if index < 2:
        INSTAGRAM_USER_MIX.append(username)
    else:
        INSTAGRAM_USER_MIX.append(user_id)
INSTAGRAM_POST_CODES = [
    "DEhf2uTJUs0",  # zuck
    "DF-rojvO4g-",  # oficialfernandatorres
    "C2SuqhGv3U0",  # crio.cafe
    "DFWFT4LyfSJ",  # ficcoesespetaculo
]
random.shuffle(INSTAGRAM_USERNAMES)
random.shuffle(INSTAGRAM_USER_IDS)
random.shuffle(INSTAGRAM_USER_MIX)
random.shuffle(INSTAGRAM_POST_CODES)


def test_instagram_profile_with_usernames(temp_dir):
    output_filename = temp_dir / "instagram-profile-1.csv"
    assert not output_filename.exists()
    collect_profile(
        client_class=Instagram,
        usernames_or_ids=INSTAGRAM_USERNAMES,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    assert len(data) == len(INSTAGRAM_USERNAMES)
    usernames_found = set(row["username"] for row in data)
    assert usernames_found == set(INSTAGRAM_USERNAMES)


def test_instagram_profile_with_user_ids(temp_dir):
    output_filename = temp_dir / "instagram-profile-2.csv"
    assert not output_filename.exists()
    collect_profile(
        client_class=Instagram,
        usernames_or_ids=INSTAGRAM_USER_IDS,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    assert len(data) == len(INSTAGRAM_USER_IDS)
    user_ids_found = set(row["id"] for row in data)
    assert user_ids_found == set(INSTAGRAM_USER_IDS)


def test_instagram_profile_with_usernames_and_user_ids(temp_dir):
    output_filename = temp_dir / "instagram-profile-3.csv"
    assert not output_filename.exists()
    collect_profile(
        client_class=Instagram,
        usernames_or_ids=INSTAGRAM_USER_MIX,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    assert len(data) == len(INSTAGRAM_USER_MIX)
    user_ids_found = set(row["id"] for row in data)
    assert user_ids_found == set(INSTAGRAM_USER_IDS)


def test_instagram_profile_posts(temp_dir):
    output_filename = temp_dir / "instagram-profile-posts-1.csv"
    assert not output_filename.exists()
    max_posts = 25
    max_posts_per_user = 10
    assert len(INSTAGRAM_USER_MIX) * max_posts_per_user > max_posts  # Guarantee the function will enforce `max_posts`
    collect_profile_posts(
        client_class=Instagram,
        usernames_or_ids=INSTAGRAM_USER_MIX,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        max_posts_per_user=max_posts_per_user,
        max_posts=max_posts,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    user_ids_found = set(row["author_id"] for row in data)
    # Test if it's a subset because user can share posts from others, so author will be another profile
    assert set(INSTAGRAM_USER_IDS).issubset(user_ids_found)
    assert len(data) == max_posts


def test_instagram_archive(temp_dir):
    output_filename = temp_dir / "instagram-archive.zip"
    assert not output_filename.exists()
    max_posts = 20
    max_posts_per_user = 5
    archive(
        client_class=Instagram,
        usernames_or_ids=INSTAGRAM_USER_MIX,
        output_zip=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        max_posts_per_user=max_posts_per_user,
        max_posts=max_posts,
        quiet=True,
    )
    assert output_filename.exists()
    zf = ZipFile(output_filename)
    filenames = [fileinfo.filename for fileinfo in zf.filelist]
    assert "profile.csv" in filenames
    assert "post.csv" in filenames
    assert sum(1 for filename in filenames if filename.startswith("media/profile_")) > 0
    assert sum(1 for filename in filenames if filename.startswith("media/post_")) > 0


def test_instagram_post(temp_dir):
    output_filename = temp_dir / "instagram-post-1.csv"
    assert not output_filename.exists()
    max_posts = 3
    assert len(INSTAGRAM_POST_CODES) > max_posts
    instagram_post(
        post_codes=INSTAGRAM_POST_CODES,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        max_posts=max_posts,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    assert len(data) == max_posts
    user_ids_found = set(row["author_id"] for row in data)
    assert len(user_ids_found) == max_posts  # There are no post codes from the same profile more than once



import datetime
import json
import re
from typing import Any, Dict, Generator, List, Union
from urllib.parse import urljoin

from .client import BaseClient
from .models import InstagramPost, InstagramProfile, Link, Location, Picture, Video
from .utils import REGEXP_LSD

REGEXP_INSTAGRAM_APP_ID = re.compile(r'"X-IG-App-ID"\s*:\s*"([^"]+)"')
REGEXP_INSTAGRAM_CSRF_TOKEN = re.compile(r'"csrf_token"\s*:\s*"([^"]+)"')
REGEXP_INSTAGRAM_PROFILE_ID = re.compile(r'"page_id"\s*:\s*"profilePage_([^"]+)"')
REGEXP_DID = re.compile("([0-9A-Z]+-[0-9A-Z]+-[0-9A-Z]+-[0-9A-Z]+-[0-9A-Z]+)")


# TODO: add username <-> user ID cache, so we avoid unneeded requests and the user don't need to manage this
class Instagram(BaseClient):
    main_url = "https://www.instagram.com/"
    graphql_url = "https://www.instagram.com/graphql/query"
    media_types_ids = {1: "photo", 2: "video", 8: "carousel"}
    media_types_names = {"XDTGraphVideo": "video", "XDTGraphSidecar": "carousel", "XDTGraphImage": "photo"}

    # TODO: implement authentication
    # TODO: list post comments

    def __post_init__(self):
        self._user_id_cache = {}
        response = self.request(
            method="GET",
            url="https://www.instagram.com/ajax/bootloader-endpoint/",
            params={
                "modules": "PolarisLoggedOutContentWallDialog.react",
                "__d": "www",
                "__user": "0",
                "__a": "1",
                "__req": "d",
            },
        )
        html = response.content.decode(response.apparent_encoding)
        values = REGEXP_DID.findall(html)
        if values:
            self.session.cookies["ig_did"] = values[0]

        response = self.request(method="GET", url=self.main_url)
        html = response.content.decode(response.apparent_encoding)
        csrf_result = REGEXP_INSTAGRAM_CSRF_TOKEN.findall(html)
        app_id_result = REGEXP_INSTAGRAM_APP_ID.findall(html)
        self.lsd = REGEXP_LSD.findall(html)[0]
        self.session.headers.update(
            {
                "x-csrftoken": csrf_result[0],
                "X-IG-App-ID": app_id_result[0],
                "x-fb-lsd": self.lsd,
                "x-asbd-id": "129477",
                "x-ig-www-claim": "0",
            }
        )

    def user_id(self, username: str) -> int:
        response = self.request(method="GET", url=urljoin(self.main_url, username))
        html = response.content.decode(response.apparent_encoding)
        return int(REGEXP_INSTAGRAM_PROFILE_ID.findall(html)[0])

    def _extract_profile(self, data):
        raw_bio_links = data.pop("bio_links", [])
        bio_links = [Link(title=link["title"], url=link["url"]) for link in raw_bio_links]
        profile_picture = Picture(url=data.pop("hd_profile_pic_url_info")["url"])
        return InstagramProfile(
            id=int(data.pop("id")),
            username=data.pop("username"),
            full_name=data.pop("full_name"),
            biography=data.pop("biography"),
            is_verified=data.pop("is_verified"),
            followers=data.pop("follower_count"),
            is_private=data.pop("is_private"),
            picture=profile_picture,
            bio_links=bio_links,
            followees=data.pop("following_count"),
            posts=data.pop("media_count"),
        )

    def profile_from_user_id(self, user_id: int, raw: bool = False) -> Union[InstagramProfile, Dict[str, Any]]:
        query_name = "PolarisProfilePageContentQuery"
        headers = {"x-fb-friendly-name": query_name}
        variables = {
            "id": str(user_id),
            "render_surface": "PROFILE",
        }
        data = {
            "lsd": self.lsd,
            "fb_api_req_friendly_name": query_name,
            "variables": json.dumps(variables),
            "doc_id": "9013883645373700",
        }
        response = self.request(method="POST", url=self.graphql_url, headers=headers, data=data)
        data = response.json()
        if "data" not in data:
            raise RuntimeError(f"Cannot get user profile: {data}")
        user = data["data"]["user"]
        if raw:
            return user
        profile = self._extract_profile(user)
        self._user_id_cache[profile.username] = profile.id
        return profile

    def profile_from_username(self, username: str, raw: bool = False) -> Union[InstagramProfile, Dict[str, Any]]:
        if username in self._user_id_cache:
            user_id = self._user_id_cache[username]
        else:
            user_id = self.user_id(username)
        return self.profile_from_user_id(user_id=user_id, raw=raw)

    def profile(self, username_or_id: Union[str, int], raw: bool = False) -> Union[InstagramProfile, Dict[str, Any]]:
        if isinstance(username_or_id, int) or username_or_id.isdigit():
            return self.profile_from_user_id(username_or_id)
        else:
            return self.profile_from_username(username_or_id)

    def _extract_media(self, node: Dict[str, Any]) -> List[Union[Picture, Video]]:
        result = []
        # Try to get the image with original size. If not found, get the biggest one
        images = list(node["image_versions2"]["candidates"])
        original_size = (node["original_width"], node["original_height"])
        selected_images = [image for image in images if (image["width"], image["height"]) == original_size]
        if selected_images:
            image = selected_images[0]
        else:
            images.sort(key=lambda obj: obj["width"], reverse=True)
            image = images[0]
        media_type = self.media_types_ids[node["media_type"]]
        if media_type == "carousel":
            media_type = "photo"
        image_media = Picture(
            url=image["url"],
            width=image["width"],
            height=image["height"],
        )
        if media_type == "video":
            video = node["video_versions"][0] if media_type == "video" else {}
            media = Video(
                type=video["type"],
                url=video["url"],
                width=video["width"],
                height=video["height"],
                thumbnail=image_media,
            )
        else:
            media = image_media
        result.append(media)
        carousel = node.get("carousel_media") or []
        for inner_node in carousel:
            result.extend(self._extract_media(inner_node))
        # TODO: extract usertags (for each media)
        return result

    def _extract_post_from_posts(self, node: Dict[str, Any]) -> InstagramPost:
        assert "taken_at" in node
        user = node["user"]
        node_pk = int(node["pk"])
        node_id = node["id"]
        timestamp = node["taken_at"]
        code = node["code"]
        if "hd_profile_pic_url_info" in user and user["hd_profile_pic_url_info"]:
            profile_picture_url = user["hd_profile_pic_url_info"]["url"]
        else:
            profile_picture_url = user["profile_pic_url"]
        author = InstagramProfile(
            id=int(user["pk"]),
            username=user["username"],
            full_name=user["full_name"],
            picture=Picture(url=profile_picture_url),
        )
        accessibility_caption = node["accessibility_caption"]
        text = (node["caption"] or {}).get("text")
        comments = node["comment_count"]
        likes = node["like_count"]
        pinned = len(node["timeline_pinned_user_ids"]) > 0
        media_type = self.media_types_ids[node["media_type"]]
        raw_location = node["location"] or {}
        location_pk = raw_location.get("pk")
        if not location_pk:
            location = None
        else:
            location = Location(
                id=int(location_pk),
                lat=raw_location["lat"],
                lng=raw_location["lng"],
                name=raw_location["name"],
            )
        published_at = datetime.datetime.fromtimestamp(timestamp).replace(tzinfo=datetime.timezone.utc)
        post_media = self._extract_media(node)
        thumbnail = None
        if media_type != "photo" and len(post_media) > 1 and isinstance(post_media[0], Picture):
            thumbnail = post_media.pop(0)
        return InstagramPost(
            id=node_id,
            pk=node_pk,
            code=code,
            type=media_type,
            pinned=pinned,
            published_at=published_at,
            text=text,
            author=author,
            comments=comments,
            likes=likes,
            accessibility_caption=accessibility_caption,
            location=location,
            thumbnail=thumbnail,
            media=post_media,
        )

    def _extract_post_media(self, node: Dict[str, Any]) -> List[Union[Picture, Video]]:
        media_type = self.media_types_names[node["__typename"]]
        images = [
            Picture(url=img["src"], width=img["config_width"], height=img["config_height"])
            for img in node["display_resources"]
        ]
        images.sort(key=lambda pic: pic.width, reverse=True)

        if media_type == "photo":
            return [images[0]]

        elif media_type == "video":
            dimensions = node["dimensions"]
            return [
                Video(
                    url=node["video_url"],
                    width=dimensions["width"],
                    height=dimensions["height"],
                    thumbnail=images[0],
                    duration=node["video_duration"],
                )
            ]

        elif media_type == "carousel":
            result = [images[0]]
            for child in node["edge_sidecar_to_children"]["edges"]:
                result.extend(self._extract_post_media(child["node"]))
            return result

    def _extract_post_from_post(self, node: Dict[str, Any]) -> InstagramPost:
        assert "taken_at_timestamp" in node
        user = node["owner"]
        node_pk = int(node["id"])
        node_id = f"{node_pk}_{user['id']}"
        timestamp = node["taken_at_timestamp"]
        code = node["shortcode"]
        author = InstagramProfile(
            id=int(user["id"]),
            username=user["username"],
            full_name=user["full_name"],
            is_verified=user["is_verified"],
            followers=user["edge_followed_by"]["count"],
            posts=user["edge_owner_to_timeline_media"]["count"],
            is_private=user["is_private"],
            picture=Picture(url=user["profile_pic_url"]),
        )
        text = node["edge_media_to_caption"]["edges"][0]["node"]["text"]
        accessibility_caption = node["accessibility_caption"]
        comments = likes = pinned = None
        media_type = self.media_types_names[node["__typename"]]
        raw_location = node["location"]
        if not raw_location:
            location = None
        else:
            json_address = raw_location["address_json"]
            address = json.loads(json_address)
            location = Location(
                id=int(raw_location["id"]),
                has_public_page=raw_location["has_public_page"],
                name=raw_location["name"],
                slug=raw_location["slug"],
                address=address["street_address"],
                zip_code=address["zip_code"],
                city=address["city_name"],
                country_code=address["country_code"],
            )
        published_at = datetime.datetime.fromtimestamp(timestamp).replace(tzinfo=datetime.timezone.utc)
        # TODO: may use `node["edge_media_to_tagged_user"]` to get tagged users

        post_media = self._extract_post_media(node)
        if media_type == "photo":
            views = plays = thumbnail = None
        elif media_type == "video":
            views = node["video_view_count"]
            plays = node["video_play_count"]
            thumbnail = Picture(url=node["thumbnail_src"])
        elif media_type == "carousel":
            views = plays = None
            thumbnail = post_media.pop(0)

        return InstagramPost(
            id=node_id,
            pk=node_pk,
            code=code,
            type=media_type,
            pinned=pinned,
            published_at=published_at,
            text=text,
            author=author,
            comments=comments,
            likes=likes,
            accessibility_caption=accessibility_caption,
            location=location,
            thumbnail=thumbnail,
            media=post_media,
            views=views,
            plays=plays,
        )

    def profile_posts_from_username(
        self, username: str, raw: bool = False
    ) -> Generator[Union[InstagramPost, Dict[str, Any]], None, None]:
        """Collect all posts from a user timeline

        Some attributes may be different according to post type:
        - "photo":
           - post.thumbnail is None
           - len(post.media) == 1
           - type(post.media[0]) is Picture
        - "video":
          - post.thumbnail is not None
          - type(post.thumbnail) is Picture
          - len(post.media) == 1
          - type(post.media[0]) is Video
        - "carousel":
          - post.thumbnail is not None
          - type(post.thumbnail) is Picture
          - len(post.media) > 1
          - type(post.media[n]) in (Picture, Video)
        """
        # TODO: check if this method gets the pinned posts
        query_name = "PolarisProfilePostsTabContentQuery_connection"
        headers = {"x-fb-friendly-name": query_name, "x-requested-with": "XMLHttpRequest"}
        cursor = None
        page_size = 12
        finished = False
        while not finished:
            variables = {
                "after": cursor,
                "before": None,
                "data": {
                    "count": page_size,
                    "include_reel_media_seen_timestamp": "true",
                    "include_relationship_info": "true",
                    "latest_besties_reel_media": "true",
                    "latest_reel_media": "true",
                },
                "first": page_size,
                "last": None,
                "username": username,
                "__relay_internal__pv__PolarisIsLoggedInrelayprovider": "true",
            }
            data = {
                "lsd": self.lsd,
                "fb_api_req_friendly_name": query_name,
                "variables": json.dumps(variables),
                "doc_id": "28967623359495231",
            }
            response = self.request(method="POST", url=self.graphql_url, headers=headers, data=data)
            response_data = response.json()
            data = response_data["data"]
            if data is None:
                return
            timeline = data["xdt_api__v1__feed__user_timeline_graphql_connection"]
            page_info = timeline["page_info"]
            edges = timeline["edges"]
            for edge in edges:
                if raw:
                    yield edge
                else:
                    post = self._extract_post_from_posts(edge["node"])
                    self._user_id_cache[post.author.username] = post.author.id
                    yield post
            finished = not page_info["has_next_page"]
            cursor = page_info["end_cursor"]

    def profile_posts_from_user_id(
        self, user_id: int, raw: bool = False
    ) -> Generator[Union[InstagramPost, Dict[str, Any]], None, None]:
        if user_id in self._user_id_cache.values():
            inverted = {value: key for key, value in self._user_id_cache.items()}
            username = inverted[user_id]
        else:
            profile = self.profile_from_user_id(user_id=user_id)
            username = profile.username
        yield from self.profile_posts_from_username(username=username, raw=raw)

    def profile_posts(
        self, username_or_id: Union[str, int], raw: bool = False
    ) -> Generator[Union[InstagramPost, Dict[str, Any]], None, None]:
        if isinstance(username_or_id, int) or username_or_id.isdigit():
            yield from self.profile_posts_from_user_id(username_or_id)
        else:
            yield from self.profile_posts_from_username(username_or_id)

    def post(self, code: str, raw: bool = False) -> InstagramPost:
        query_name = "PolarisPostActionLoadPostQueryQuery"
        headers = {"x-fb-friendly-name": query_name}
        variables = {
            "shortcode": code,
            "fetch_tagged_user_count": None,
            "hoisted_comment_id": None,
            "hoisted_reply_id": None,
        }
        data = {
            "lsd": self.lsd,
            "fb_api_req_friendly_name": query_name,
            "fb_api_caller_class": "RelayModern",
            "variables": json.dumps(variables),
            "doc_id": "8845758582119845",
        }
        response = self.request(method="POST", url=self.graphql_url, headers=headers, data=data)
        data = response.json()
        node = data["data"]["xdt_shortcode_media"]
        if raw:
            return node
        post = self._extract_post_from_post(node)
        self._user_id_cache[post.author.username] = post.author.id
        return post



import datetime
import io
import json
import subprocess
from dataclasses import asdict, dataclass, field, fields
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Type, Union


def flatten(row: Dict[str, Any], nested_fields: Dict[str, Any]):
    """Flatten row inner dictionaries inline"""
    for field_name, FieldClass in nested_fields.items():
        inner_values = row.pop(field_name)
        if not inner_values:
            row.update(
                {
                    f"{field_name}_{inner_field.name}": None
                    for inner_field in fields(FieldClass)
                    if inner_field.name != "content"
                }
            )
        else:
            for key, value in inner_values.items():
                if key != "content":
                    row[f"{field_name}_{key}"] = value


class BaseModel:
    def serialize(self) -> Dict[str, Any]:
        """Converts dataclass values into a `dict` (`content` won't be serialized!)"""
        row = asdict(self)
        if "content" in row:
            del row["content"]
        return row


class BaseMedia(BaseModel):

    def get_media(self) -> List[Type["BaseMedia"]]:
        """Return all media objects of this object"""
        raise NotImplementedError("BaseMedia child class must implement `get_media` method")

    def save(self, filename: Union[Path, str]):
        if self.content is None:
            raise ValueError("Object has no content (download first)")
        filename = Path(filename)
        filename.parent.mkdir(parents=True, exist_ok=True)
        filename.write_bytes(self.content)


class BaseProfile(BaseModel):

    def get_media(self) -> List[Type[BaseMedia]]:
        """Return all media objects of this profile (not from its posts)"""
        raise NotImplementedError("Profile child class must implement `get_media` method")


class BasePost(BaseModel):

    def get_media(self) -> List[Type[BaseMedia]]:
        """Return all media objects of this post"""
        raise NotImplementedError("Post child class must implement `get_media` method")


# Generic classes used by both sites


@dataclass
class Picture(BaseMedia):
    url: str = field(repr=False)
    width: Optional[int] = None  # InstagramProfile doesn't have it
    height: Optional[int] = None  # InstagramProfile doesn't have it
    content: Optional[bytes] = field(repr=False, default=None)
    content_type: Optional[str] = None

    def get_media(self) -> List[Type[BaseMedia]]:
        """Return all media objects related to this picture (just itself)"""
        return [self]

    def calculate_size(self):
        """Fill `width` and `height` fields by reading image content. Requires a pre-downloaded picture and `Pillow`"""
        from PIL import Image  # noqa

        if self.content is None:
            raise ValueError(
                "Cannot calculate size for an image without content - download it first with `<ClientClass>.download(picture)`"
            )
        img = Image.open(io.BytesIO(self.content))
        self.width, self.height = img.size


@dataclass
class Video(BaseMedia):
    url: str = field(repr=False)
    type: Optional[int] = None
    width: Optional[int] = None
    height: Optional[int] = None
    thumbnail: Optional[Picture] = None
    duration: Optional[float] = None
    content: Optional[bytes] = field(repr=False, default=None)
    content_type: Optional[str] = None

    def serialize(self) -> Dict[str, Any]:
        row = super().serialize()
        flatten(row, {"thumbnail": Picture})
        return row

    def get_media(self) -> List[Type[BaseMedia]]:
        """Return all media objects related to this video (itself and the thumbnail, if avaiable)"""
        result = [self]
        if self.thumbnail:
            result.append(self.thumbnail)
        return result

    def calculate_duration(self):
        """Fill `duration` field by reading video content. Requires a pre-downloaded video and `ffmpeg`"""
        if self.content is None:
            raise ValueError(
                "Cannot calculate duration for a video without content - download it first with `<ClientClass>.download(video)`"
            )
        cmd = [
            "ffprobe",
            "-v",
            "quiet",
            "-print_format",
            "json",
            "-show_format",
            "-show_streams",
            "-",
        ]
        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate(input=self.content)
        if stderr:
            raise RuntimeError(f"Error: checking video duration {stderr.decode('utf-8')}")
        data = json.loads(stdout)
        streams_duration = [float(stream["duration"]) for stream in data["streams"]]
        if streams_duration:
            self.duration = max(streams_duration)
        else:
            value = data["format"].get("duration", None)
            if value is None:
                raise RuntimeError("Cannot get duration from ffprobe: data not found")
            self.duration = float(value)


@dataclass
class Link(BaseModel):
    url: str
    title: str = None  # Used in InstagramProfile
    is_verified: bool = None  # Used in Threads
    id: int = None  # Used in Threads
    display_text: str = None  # Used in Threads


@dataclass
class Location(BaseModel):
    id: int
    name: str
    slug: Optional[str] = None
    lat: Optional[float] = None
    lng: Optional[float] = None
    has_public_page: Optional[bool] = None
    address: Optional[str] = None
    zip_code: Optional[str] = None
    city: Optional[str] = None
    country_code: Optional[str] = None


# Threads-related dataclasses


@dataclass
class ThreadsProfile(BaseProfile):
    id: int
    username: str
    full_name: str
    biography: str
    is_verified: bool
    followers: int
    pictures: List[Picture]
    is_private: bool
    bio_links: Optional[List[Link]] = None
    is_threads_only_user: Optional[bool] = None

    @property
    def picture(self) -> Picture:
        return self.pictures[0]

    def serialize(self) -> Dict[str, Any]:
        row = super().serialize()
        row["bio_links"] = "\n".join(link["url"] for link in row["bio_links"])
        pictures = row.pop("pictures")
        pictures.sort(key=lambda obj: obj["width"], reverse=True)
        if not pictures:
            row["picture_url"] = row["picture_width"] = row["picture_height"] = None
        else:
            row["picture_url"] = pictures[0]["url"]
            row["picture_width"] = pictures[0]["width"]
            row["picture_height"] = pictures[0]["height"]
        return row

    @property
    def url(self):
        return f"https://www.threads.net/@{self.userame}"

    def get_media(self) -> List[Picture]:
        """Return all media objects of this profile (not from its posts)"""
        return self.pictures


@dataclass
class ThreadsPost(BasePost):
    id: str
    user_id: str
    username: str
    is_verified: bool
    text: str
    links: List[Link]
    published_at: datetime.datetime
    likes: int
    replies: int
    reposts: int
    quotes: int
    is_private: bool
    media: List[Union[Picture, Video]] = field(default_factory=list)
    reply_control: Optional[str] = None
    media_type: Optional[int] = None
    accessibility_caption: Optional[str] = None
    is_paid_partnership: Optional[bool] = None
    like_and_view_counts_disabled: Optional[bool] = None
    has_audio: Optional[bool] = None
    original_width: Optional[int] = None
    original_height: Optional[int] = None
    code: Optional[str] = None
    reshares: Optional[int] = None

    def serialize(self) -> Dict[str, Any]:
        row = super().serialize()
        for key in ("links", "media"):
            row[key] = "\n".join(link["url"] for link in row[key])
        return row

    @property
    def url(self):
        return f"https://www.threads.net/@{self.userame}/post/{self.code}"

    def get_media(self) -> List[Type[BaseMedia]]:
        """Return all media objects of this post"""
        return self.media or []


# Instagram-related dataclasses


@dataclass
class InstagramProfile(BaseProfile):
    id: int
    username: str
    full_name: str
    picture: Picture
    is_verified: Optional[bool] = None
    followers: Optional[int] = None
    followees: Optional[int] = None
    posts: Optional[int] = None
    is_private: Optional[bool] = None
    biography: Optional[str] = None
    bio_links: Optional[List[Link]] = None

    def serialize(self) -> Dict[str, Any]:
        row = super().serialize()
        flatten(row, {"picture": Picture})
        row["bio_links"] = "\n".join(link["url"] for link in row["bio_links"])
        return row

    @property
    def url(self):
        return f"https://www.instagram.com/{self.userame}/"

    def get_media(self) -> List[Picture]:
        """Return all media objects of this profile (not from its posts)"""
        return [self.picture]


@dataclass
class InstagramPost(BasePost):
    id: str
    pk: int
    code: str
    type: Union[Literal["photo"], Literal["video"], Literal["carousel"]]
    published_at: datetime.datetime
    author: InstagramProfile
    media: List[Union[Picture, Video]]
    pinned: Optional[bool] = None
    comments: Optional[int] = None
    likes: Optional[int] = None
    views: Optional[int] = None
    plays: Optional[int] = None
    thumbnail: Optional[Picture] = None
    text: Optional[str] = None
    author_full_name: Optional[str] = None
    accessibility_caption: Optional[str] = None
    location: Optional[Location] = None

    @property
    def url(self):
        return f"https://www.instagram.com/{self.author_username}/{self.code}/"

    def serialize(self) -> Dict[str, Any]:
        row = super().serialize()
        flatten(row, {"author": InstagramProfile, "thumbnail": Picture, "location": Location})
        row["media"] = "\n".join(obj["url"] for obj in row["media"])
        row["author_picture_url"] = row.pop("author_picture")["url"]
        return row

    def get_media(self) -> List[Type[BaseMedia]]:
        """Return all media objects of this post"""
        return ([self.thumbnail] if self.thumbnail else []) + (self.media or [])



import argparse
import csv
import io
from pathlib import Path
from typing import Any, List, Optional, Union
from zipfile import ZipFile

from . import __version__
from .instagram import Instagram
from .threads import Threads

# TODO: implement threads-archive


class ProgressBar:
    """Simple and easy-to-user progress bar for the command-line interface"""

    def __init__(self, total: Optional[int] = None, title: Optional[str] = None):
        self.last_line_length = None
        self.title = title
        self.total = total

    def update(
        self,
        current: int,
        done: bool = False,
        total: Optional[int] = None,
        title: Optional[str] = None,
        suffix: Optional[str] = None,
        end="",
    ):
        if total is None:
            total = self.total
        if not done:  # The `current` item was started but not done yet
            percent = 100 * (current - 1 if current != 0 else 0) / total
        else:
            percent = (100 * current) / total
        title_str = title or self.title or ""
        suffix_str = suffix if suffix is not None else ""
        line = f"\r{title_str}{current:03d}/{total:03d} ({percent:3.2f}%){suffix_str}"
        if self.last_line_length is not None and self.last_line_length > len(line):
            print("\r" + " " * self.last_line_length, end="")
        print(line, flush=True, end=end)
        self.last_line_length = len(line)


class QuietProgressBar(ProgressBar):
    """A ProgressBar that does nothing, but has the same API"""

    def update(
        self,
        current: int,
        done: bool = False,
        total: Optional[int] = None,
        title: Optional[str] = None,
        suffix: Optional[str] = None,
        end="",
    ):
        pass


def collect_profile(
    client_class: Any,
    usernames_or_ids: List[Union[str, int]],
    output_csv: Path,
    min_wait: float,
    max_wait: float,
    quiet: bool = False,
):
    client = client_class(min_wait=min_wait, max_wait=max_wait)
    output_csv.parent.mkdir(parents=True, exist_ok=True)
    with output_csv.open(mode="w") as fobj:
        writer = None
        ProgressClass = ProgressBar if not quiet else QuietProgressBar
        progress = ProgressClass(title=f"{client_class.__name__} Profile ", total=len(usernames_or_ids))
        for index, username_or_id in enumerate(usernames_or_ids, start=1):
            progress.update(index, done=False)
            profile = client.profile(username_or_id)
            row = profile.serialize()
            if writer is None:
                writer = csv.DictWriter(fobj, fieldnames=list(row.keys()))
                writer.writeheader()
            writer.writerow(row)
            progress.update(index, done=True)
        progress.update(index, done=True, end="\n")


def collect_profile_posts(
    client_class: Any,
    usernames_or_ids: List[Union[str, int]],
    output_csv: Path,
    min_wait: float,
    max_wait: float,
    max_posts_per_user: Optional[int] = None,
    max_posts: Optional[int] = None,
    quiet: bool = False,
):
    client = client_class(min_wait=min_wait, max_wait=max_wait)
    output_csv.parent.mkdir(parents=True, exist_ok=True)
    with output_csv.open(mode="w") as fobj:
        writer = None
        ProgressClass = ProgressBar if not quiet else QuietProgressBar
        progress = ProgressClass(title=f"{client_class.__name__} Profile ", total=len(usernames_or_ids))
        downloaded_posts = 0
        finished = False
        for index, username_or_id in enumerate(usernames_or_ids, start=1):
            progress.update(index, done=False)
            for post_index, post in enumerate(client.profile_posts(username_or_id), start=1):
                row = post.serialize()
                if writer is None:
                    writer = csv.DictWriter(fobj, fieldnames=list(row.keys()))
                    writer.writeheader()
                writer.writerow(row)
                progress.update(index, suffix=f" Post {post_index:03d}", done=False)
                downloaded_posts += 1
                if max_posts is not None and downloaded_posts == max_posts:
                    finished = True
                    break
                elif max_posts_per_user is not None and post_index == max_posts_per_user:
                    break
            progress.update(index, done=True)
            if finished:
                break
        progress.update(index, done=True, end="\n")


def archive(
    client_class: Any,
    usernames_or_ids: List[Union[str, int]],
    output_zip: Path,
    min_wait: float,
    max_wait: float,
    max_posts_per_user: Optional[int] = None,
    max_posts: Optional[int] = None,
    quiet: bool = False,
):
    client = client_class(min_wait=min_wait, max_wait=max_wait)
    zf = ZipFile(output_zip, mode="w")

    profiles = []
    ProgressClass = ProgressBar if not quiet else QuietProgressBar
    progress = ProgressClass(title=f"{client_class.__name__} Profile ", total=len(usernames_or_ids))
    with io.TextIOWrapper(zf.open("profile.csv", mode="w"), encoding="utf-8") as fobj:
        writer = None
        for index, username_or_id in enumerate(usernames_or_ids, start=1):
            progress.update(index, done=False)
            profile = client.profile(username_or_id)
            profiles.append(profile)
            row = profile.serialize()
            if writer is None:
                writer = csv.DictWriter(fobj, fieldnames=list(row.keys()))
                writer.writeheader()
            writer.writerow(row)
            progress.update(index, done=True)
        progress.update(index, done=True, end="\n")

    progress = ProgressClass(title=f"{client_class.__name__} Profile media (download) ", total=len(profiles))
    progress.update(0, done=True)
    for index, media in enumerate(client.download_many(profiles), start=1):
        progress.update(index, done=True)
    progress.update(index, done=True, end="\n")

    progress = ProgressClass(title=f"{client_class.__name__} Profile media (save) ", total=len(profiles))
    for index, profile in enumerate(profiles, start=1):
        progress.update(index, done=False)
        extension = profile.picture.content_type.split("/")[-1].lower()
        extension = {"jpeg": "jpg"}.get(extension, extension)
        with zf.open(f"media/profile_{profile.id}.{extension}", mode="w") as fobj:
            fobj.write(profile.picture.content)
        progress.update(index, done=True)
    progress.update(index, done=True, end="\n")

    posts = []
    total_post_media = 0
    with io.TextIOWrapper(zf.open("post.csv", mode="w"), encoding="utf-8") as fobj:
        writer = None
        progress = ProgressClass(title=f"{client_class.__name__} Profile posts ", total=len(profiles))
        downloaded_posts = 0
        finished = False
        for index, profile in enumerate(profiles, start=1):
            progress.update(index, done=False)
            for post_index, post in enumerate(client.profile_posts(profile.username), start=1):
                posts.append(post)
                total_post_media += len(post.get_media())
                row = post.serialize()
                if writer is None:
                    writer = csv.DictWriter(fobj, fieldnames=list(row.keys()))
                    writer.writeheader()
                writer.writerow(row)
                progress.update(index, suffix=f" Post {post_index:03d}", done=False)
                downloaded_posts += 1
                if max_posts is not None and downloaded_posts == max_posts:
                    finished = True
                    break
                elif max_posts_per_user is not None and post_index == max_posts_per_user:
                    break
            progress.update(index, done=True)
            if finished:
                break
        progress.update(index, done=True, end="\n")

    progress = ProgressClass(title=f"{client_class.__name__} Profile posts media (download) ", total=total_post_media)
    progress.update(0, done=True)
    for index, media in enumerate(client.download_many(posts), start=1):
        progress.update(index, done=True)
    progress.update(index, done=True, end="\n")

    progress = ProgressClass(title=f"{client_class.__name__} Profile posts media (save) ", total=total_post_media)
    counter = 0
    for post in posts:
        for post_media_index, media in enumerate(post.get_media(), start=1):
            counter += 1
            progress.update(counter, done=False)
            extension = media.content_type.split("/")[-1].lower()
            extension = {"jpeg": "jpg"}.get(extension, extension)
            with zf.open(f"media/post_{post.id}_{post_media_index:03d}.{extension}", mode="w") as fobj:
                fobj.write(media.content)
            progress.update(counter, done=True)
    progress.update(counter, done=True, end="\n")

    zf.close()


def instagram_post(
    post_codes: List[str],
    output_csv: Path,
    min_wait: float,
    max_wait: float,
    max_posts: Optional[int] = None,
    quiet: bool = False,
):
    client = Instagram(min_wait=min_wait, max_wait=max_wait)
    output_csv.parent.mkdir(parents=True, exist_ok=True)
    with output_csv.open(mode="w") as fobj:
        writer = None
        ProgressClass = ProgressBar if not quiet else QuietProgressBar
        progress = ProgressClass(title="Post ", total=len(post_codes))
        progress.update(0, done=False)
        for index, post_code in enumerate(post_codes, start=1):
            post = client.post(post_code)
            row = post.serialize()
            if writer is None:
                writer = csv.DictWriter(fobj, fieldnames=list(row.keys()))
                writer.writeheader()
            writer.writerow(row)
            progress.update(index, done=True)
            if max_posts is not None and index == max_posts:
                break
        progress.update(index, done=True, end="\n")


def main():
    MAX_POSTS = 96
    MAX_POSTS_PER_USER = 48
    class_map = {"instagram": Instagram, "threads": Threads}

    parser = argparse.ArgumentParser(prog="outgram")
    parser.add_argument("--version", action="version", version=f"%(prog)s {'.'.join(map(str, __version__))}")
    parser.add_argument(
        "--min-wait", "-t", type=float, default=0.5, help="Minimum time (in seconds) to wait between requests"
    )
    parser.add_argument(
        "--max-wait", "-T", type=float, default=2.0, help="Maximum time (in seconds) to wait between requests"
    )
    subparsers = parser.add_subparsers(dest="network", required=True)

    instagram_parser = subparsers.add_parser("instagram")
    threads_parser = subparsers.add_parser("threads")
    network_subparsers = {
        "instagram": instagram_parser.add_subparsers(dest="action", required=True),
        "threads": threads_parser.add_subparsers(dest="action", required=True),
    }

    for network in class_map.keys():
        network_subparser = network_subparsers[network]
        parser_profile = network_subparser.add_parser(
            "profile",
            description="Get information regarding one or more profiles. Only the biggest picture URL is exported. Only the URLs of bio_links are exported (separated by \\n)",
        )
        parser_profile.add_argument("--quiet", "-q", action="store_true", help="Do not show a progress bar")
        parser_profile.add_argument("username_or_id", nargs="+")
        parser_profile.add_argument("output_csv", type=Path)

        parser_profile_posts = network_subparser.add_parser(
            "profile-posts",
            description="Get latests posts of a list of profiles. A list of usernames or user IDs can be passed (it takes one more request per username, since we need to figure out the user ID). Only the URLs of links, pictures and videos are exported (separated by \\n)",
        )
        parser_profile_posts.add_argument("--quiet", "-q", action="store_true", help="Do not show a progress bar")
        parser_profile_posts.add_argument(
            "--max-posts-per-user", "-m", type=int, default=MAX_POSTS_PER_USER, help="Maximum posts to get per user"
        )
        parser_profile_posts.add_argument(
            "--max-posts", "-M", type=int, default=MAX_POSTS, help="Maximum posts to get (total)"
        )
        parser_profile_posts.add_argument("username_or_id", nargs="+")
        parser_profile_posts.add_argument("output_csv", type=Path)

        parser_archive = network_subparser.add_parser(
            "archive",
            description="Get all posts and media metadata and download media for a list of profiles and save into a ZIP file. A list of usernames or user IDs can be passed.",
        )
        parser_archive.add_argument("--quiet", "-q", action="store_true", help="Do not show a progress bar")
        parser_archive.add_argument(
            "--max-posts-per-user", "-m", type=int, default=MAX_POSTS_PER_USER, help="Maximum posts to get per user"
        )
        parser_archive.add_argument(
            "--max-posts", "-M", type=int, default=MAX_POSTS, help="Maximum posts to get (total)"
        )
        parser_archive.add_argument("username_or_id", nargs="+")
        parser_archive.add_argument("output_zip", type=Path)

    parser_instagram_post = network_subparsers["instagram"].add_parser(
        "post",
        description="Get posts from a list of post codes (the code is found in the URL: <https://www.instagram.com/p/CODE/>).",
    )
    parser_instagram_post.add_argument("--quiet", "-q", action="store_true", help="Do not show a progress bar")
    parser_instagram_post.add_argument(
        "--max-posts", "-M", type=int, default=MAX_POSTS, help="Maximum posts to get (total)"
    )
    parser_instagram_post.add_argument("post_code", nargs="+")
    parser_instagram_post.add_argument("output_csv", type=Path)

    args = parser.parse_args()

    if args.action == "profile":
        collect_profile(
            client_class=class_map[args.network],
            usernames_or_ids=args.username_or_id,
            output_csv=args.output_csv,
            min_wait=args.min_wait,
            max_wait=args.max_wait,
            quiet=args.quiet,
        )

    elif args.action == "profile-posts":
        collect_profile_posts(
            client_class=class_map[args.network],
            usernames_or_ids=args.username_or_id,
            output_csv=args.output_csv,
            min_wait=args.min_wait,
            max_wait=args.max_wait,
            max_posts_per_user=args.max_posts_per_user,
            max_posts=args.max_posts,
            quiet=args.quiet,
        )

    elif args.action == "archive":
        archive(
            client_class=class_map[args.network],
            usernames_or_ids=args.username_or_id,
            output_zip=args.output_zip,
            min_wait=args.min_wait,
            max_wait=args.max_wait,
            max_posts_per_user=args.max_posts_per_user,
            max_posts=args.max_posts,
            quiet=args.quiet,
        )

    elif (args.network, args.action) == ("instagram", "post"):
        instagram_post(
            post_codes=args.post_code,
            output_csv=args.output_csv,
            min_wait=args.min_wait,
            max_wait=args.max_wait,
            max_posts=args.max_posts,
            quiet=args.quiet,
        )

    # TODO: implement threads-post


if __name__ == "__main__":
    main()



import datetime
import json
from typing import Any, Dict, Generator, Union

from .client import BaseClient
from .models import Link, Picture, ThreadsPost, ThreadsProfile, Video
from .utils import REGEXP_LSD


# TODO: add username <-> user ID cache, so we avoid unneeded requests and the user don't need to manage this
class Threads(BaseClient):
    main_url = "https://www.threads.net/"
    graphql_url = "https://www.threads.net/api/graphql"

    # TODO: implement authentication
    # TODO: list post comments (needs authentication)

    def __post_init__(self):
        response = self.request(method="GET", url=self.main_url)
        html = response.content.decode(response.apparent_encoding)
        self.session.headers["x-fb-lsd"] = REGEXP_LSD.findall(html)[0]
        self._user_id_cache = {}

    def _extract_profile(self, json_data: dict) -> ThreadsProfile:
        user = json_data["data"]["user"]
        pictures = []
        if "profile_pic_url" in user:
            pictures.append(Picture(url=user["profile_pic_url"], width=150, height=150))  # TODO: check hard-coded size
        pictures.extend(
            [
                Picture(url=pic["url"], width=pic["width"], height=pic["height"])
                for pic in user.get("hd_profile_pic_versions", [])
            ]
        )
        # TODO: bio_links not working?
        if "bio_links" in user:
            bio_links = [
                Link(url=link["url"], is_verified=link["is_verified"], id=int(link["link_id"]))
                for link in user.get("bio_links", [])
            ]
        else:
            bio_links = []
        return ThreadsProfile(
            id=int(user["id"]),
            username=user["username"],
            full_name=user["full_name"],
            biography=user["biography"],
            is_verified=user["is_verified"],
            followers=user["follower_count"],
            pictures=pictures,
            is_private=user["text_post_app_is_private"],
            bio_links=bio_links,
            is_threads_only_user=user.get("is_threads_only_user"),
        )

    def simple_profile(self, username: str, raw: bool = False) -> Union[ThreadsProfile, Dict[str, Any]]:
        """Get basic profile information. For complete info, use `.full_profile`"""
        query_name = "BarcelonaUsernameHoverCardImplQuery"
        headers = {
            "x-fb-friendly-name": query_name,
        }
        query_variables = {
            "username": username,
            "__relay_internal__pv__BarcelonaShouldShowFediverseM075Featuresrelayprovider": "false",
        }
        data = {
            "fb_api_req_friendly_name": query_name,
            "variables": json.dumps(query_variables),
            "lsd": self.session.headers["x-fb-lsd"],
            "doc_id": "7679337195500348",
        }
        response = self.request(method="POST", url=self.graphql_url, headers=headers, data=data)
        data = response.json()
        if raw:
            return data
        profile = self._extract_profile(data)
        self._user_id_cache[profile.username] = profile.id
        return profile

    def profile_from_user_id(self, user_id: int, raw: bool = False) -> Union[ThreadsProfile, Dict[str, Any]]:
        """Get full profile info for an user. If you only have the username, get the ID with `simple_profile()`"""
        query_name = "BarcelonaProfilePageQuery"
        headers = {
            "x-fb-friendly-name": query_name,
        }
        query_variables = {
            "userID": user_id,
            "__relay_internal__pv__BarcelonaIsLoggedInrelayprovider": "false",
            "__relay_internal__pv__BarcelonaIsPASTISWarningEnabledrelayprovider": "false",
            "__relay_internal__pv__BarcelonaIsSableEnabledrelayprovider": "false",
            "__relay_internal__pv__BarcelonaIsLinkVerificationEnabledrelayprovider": "false",
            "__relay_internal__pv__BarcelonaShouldShowFediverseM075Featuresrelayprovider": "false",
        }
        data = {
            "fb_api_req_friendly_name": query_name,
            "variables": json.dumps(query_variables),
            "lsd": self.session.headers["x-fb-lsd"],
            "doc_id": "8454848744553415",
        }
        response = self.request(method="POST", url=self.graphql_url, headers=headers, data=data)
        data = response.json()
        return self._extract_profile(data) if not raw else data  # `raw` is useful for debugging purposes

    def profile_from_username(self, username: str, raw: bool = False) -> Union[ThreadsProfile, Dict[str, Any]]:
        if username in self._user_id_cache:
            user_id = self._user_id_cache[username]
        else:
            simple_profile = self.simple_profile(username)
            user_id = simple_profile.id
        return self.profile_from_user_id(user_id=user_id, raw=raw)

    def profile(self, username_or_id: Union[str, int], raw: bool = False) -> Union[ThreadsProfile, Dict[str, Any]]:
        if isinstance(username_or_id, int) or username_or_id.isdigit():
            return self.profile_from_user_id(username_or_id)
        else:
            return self.profile_from_username(username_or_id)

    def _extract_post(self, post_data: dict) -> ThreadsPost:
        if len(post_data.get("carousel_media") or []) > 0:
            inner_media = post_data["carousel_media"]
        else:
            inner_media = [post_data]
        post_media = []
        for obj in inner_media:
            obj_picture = None
            if "image_versions2" in obj and "candidates" in obj["image_versions2"]:
                images = [
                    Picture(pic["url"], pic["width"], pic["height"]) for pic in obj["image_versions2"]["candidates"]
                ]
                images.sort(key=lambda obj: obj.width, reverse=True)
                if images:
                    obj_picture = images[0]
            if len(obj.get("video_versions") or []) > 0:
                video = obj["video_versions"][0]
                video = Video(type=video["type"], url=video["url"])
                if obj_picture is not None:
                    video.thumbnail = obj_picture
                post_media.append(video)
            elif obj_picture is not None:
                post_media.append(obj_picture)

        text = ""
        links = []
        if "text_post_app_info" in post_data and "text_fragments" in post_data["text_post_app_info"]:
            for fragment in post_data["text_post_app_info"]["text_fragments"]["fragments"]:
                text += fragment["plaintext"]
                if fragment["fragment_type"] == "link" and fragment["link_fragment"]:
                    links.append(
                        Link(
                            url=fragment["link_fragment"]["uri"],
                            display_text=fragment["link_fragment"]["display_text"],
                        )
                    )
        user = post_data["user"]
        return ThreadsPost(
            id=post_data["id"],
            user_id=int(user["pk"]),
            username=user["username"],
            is_verified=user["is_verified"],
            text=text,
            links=links,
            published_at=datetime.datetime.fromtimestamp(post_data["taken_at"]).replace(tzinfo=datetime.timezone.utc),
            likes=post_data["like_count"],
            replies=post_data["text_post_app_info"].get("direct_reply_count", 0),
            reposts=post_data["text_post_app_info"].get("repost_count", 0),
            quotes=post_data["text_post_app_info"].get("quote_count", 0),
            is_private=user.get("text_post_app_is_private", False),
            media=post_media,
            reply_control=post_data["text_post_app_info"].get("reply_control"),
            media_type=post_data.get("media_type"),
            accessibility_caption=post_data.get("accessibility_caption"),
            is_paid_partnership=post_data.get("is_paid_partnership"),
            like_and_view_counts_disabled=post_data.get("like_and_view_counts_disabled"),
            has_audio=post_data.get("has_audio"),
            original_width=post_data.get("original_width"),
            original_height=post_data.get("original_height"),
            code=post_data.get("code"),
            reshares=post_data["text_post_app_info"].get("reshare_count"),
        )

    def profile_posts_from_user_id(
        self, user_id: int, raw: bool = False
    ) -> Generator[Union[ThreadsPost, Dict[str, Any]], None, None]:
        """
        Get last posts for user with ID `user_id`

        Call `profile(username)` or `full_profile(username)` to get `user_id` from a username
        """
        query_name = "BarcelonaProfileThreadsTabQuery"
        headers = {
            "x-fb-friendly-name": query_name,
        }
        query_variables = {
            "userID": user_id,
            "__relay_internal__pv__BarcelonaIsLoggedInrelayprovider": "false",
            "__relay_internal__pv__BarcelonaIsInlineReelsEnabledrelayprovider": "true",
            "__relay_internal__pv__BarcelonaOptionalCookiesEnabledrelayprovider": "true",
            "__relay_internal__pv__BarcelonaShowReshareCountrelayprovider": "true",
            "__relay_internal__pv__BarcelonaQuotedPostUFIEnabledrelayprovider": "false",
            "__relay_internal__pv__BarcelonaIsCrawlerrelayprovider": "false",
            "__relay_internal__pv__BarcelonaShouldShowFediverseM075Featuresrelayprovider": "false",
        }
        data = {
            "fb_api_req_friendly_name": query_name,
            "variables": json.dumps(query_variables),
            "lsd": self.session.headers["x-fb-lsd"],
            "doc_id": "27224795300468294",
        }
        response = self.request(method="POST", url=self.graphql_url, headers=headers, data=data)
        data = response.json()
        for post_data in data["data"]["mediaData"]["edges"]:
            raw_post = post_data["node"]["thread_items"][0]["post"]
            if raw:
                yield raw_post
            else:
                post = self._extract_post(raw_post)
                self._user_id_cache[post.username] = post.user_id
                yield post
        # TODO: if logged in, get `end_cursor` inside `page_info` and paginate

    def profile_posts_from_username(
        self, username: str, raw: bool = False
    ) -> Generator[Union[ThreadsPost, Dict[str, Any]], None, None]:
        if username in self._user_id_cache:
            user_id = self._user_id_cache[username]
        else:
            simple_profile = self.simple_profile(username)
            user_id = simple_profile.id
        yield from self.profile_posts_from_user_id(user_id=user_id, raw=raw)

    def profile_posts(
        self, username_or_id: Union[str, int], raw: bool = False
    ) -> Generator[Union[ThreadsPost, Dict[str, Any]], None, None]:
        if isinstance(username_or_id, int) or username_or_id.isdigit():
            yield from self.profile_posts_from_user_id(username_or_id)
        else:
            yield from self.profile_posts_from_username(username_or_id)



import time
from multiprocessing.pool import ThreadPool
from typing import Generator, List, Optional, Type, Union

import requests

from .models import BaseMedia, BasePost, BaseProfile
from .utils import default_browser_headers, random_user_agent, random_waiter


class BaseClient:

    def __init__(self, user_agent: Optional[str] = None, min_wait: float = 3.5, max_wait: float = 5.0):
        if user_agent is None:
            user_agent = random_user_agent()
        self._last_request_time = None
        self._wait = random_waiter(min_wait, max_wait)
        self.session = requests.Session()
        self.session.headers.update(default_browser_headers(user_agent))
        self.__post_init__()

    def __post_init__(self):
        raise NotImplementedError("Child classes must implement `__post_init__` method")

    def request(self, method, url, params=None, headers=None, data=None):
        if self._last_request_time is not None:  # Do not wait before the first request
            self._wait(self._last_request_time)
        response = self.session.request(method=method, url=url, params=params, headers=headers, data=data)
        self._last_request_time = time.time()  # TODO: this may not be thread safe
        return response

    def _download_media(self, media: Type[BaseMedia]) -> Type[BaseMedia]:
        """Download from the main URL of a media object (picture or video)"""
        response = self.request("GET", media.url)
        media.content = response.content
        media.content_type = response.headers["Content-Type"]
        return media

    def download_many(
        self, objects: List[Union[Type[BaseMedia], Type[BaseProfile], Type[BasePost]]], parallel: int = 8
    ) -> Generator[Type[BaseMedia], None, None]:
        """Download related media (skips already downloaded). Yields each one when download is finished and also updates inline"""
        to_download = []
        for obj in objects:
            to_download.extend([media for media in obj.get_media() if media.content is None])
        if len(to_download) == 1:  # Do not use threads
            yield self._download_media(to_download[0])
        else:
            with ThreadPool(processes=parallel) as pool:
                for media in pool.imap_unordered(self._download_media, to_download):
                    yield media

    def download(
        self, obj: Union[Type[BaseMedia], Type[BaseProfile], Type[BasePost]], parallel: int = 4
    ) -> Generator[Type[BaseMedia], None, None]:
        """Download related media (skips already downloaded). Yields each one when download is finished and also updates inline"""
        yield from self.download_many(objects=[obj], parallel=parallel)



"""End-to-end tests for the command-line interface main functions, which will run most part of the codebase"""

import csv
import random
from collections import Counter
from zipfile import ZipFile

from outgram.cli import archive, collect_profile, collect_profile_posts
from outgram.threads import Threads

MIN_WAIT = 5.0
MAX_WAIT = 10.0
THREADS_USERNAME_ID = {
    "ayubionet": "5633396222",
    "diogocortiz": "197762208",
    "wsj": "18133069",
    "nonoinvestidor": "42799100757",
    "filosofia.liquida": "6828796459",
    "mulheresdocafemataotomazina": "3540507650",
}
THREADS_USERNAMES = list(THREADS_USERNAME_ID.keys())
THREADS_USER_IDS = list(THREADS_USERNAME_ID.values())
THREADS_USER_MIX = []
for index, (username, user_id) in enumerate(THREADS_USERNAME_ID.items()):
    if index < 3:
        THREADS_USER_MIX.append(username)
    else:
        THREADS_USER_MIX.append(user_id)
random.shuffle(THREADS_USERNAMES)
random.shuffle(THREADS_USER_IDS)
random.shuffle(THREADS_USER_MIX)


def test_threads_profile_with_usernames(temp_dir):
    output_filename = temp_dir / "threads-profile-1.csv"
    assert not output_filename.exists()
    collect_profile(
        client_class=Threads,
        usernames_or_ids=THREADS_USERNAMES,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    assert len(data) == len(THREADS_USERNAMES)
    usernames_found = set(row["username"] for row in data)
    assert usernames_found == set(THREADS_USERNAMES)


def test_threads_profile_with_user_ids(temp_dir):
    output_filename = temp_dir / "threads-profile-2.csv"
    assert not output_filename.exists()
    collect_profile(
        client_class=Threads,
        usernames_or_ids=THREADS_USER_IDS,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    assert len(data) == len(THREADS_USER_IDS)
    user_ids_found = set(row["id"] for row in data)
    assert user_ids_found == set(THREADS_USER_IDS)


def test_threads_profile_with_usernames_and_user_ids(temp_dir):
    output_filename = temp_dir / "threads-profile-3.csv"
    assert not output_filename.exists()
    collect_profile(
        client_class=Threads,
        usernames_or_ids=THREADS_USER_MIX,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    assert len(data) == len(THREADS_USER_MIX)
    user_ids_found = set(row["id"] for row in data)
    assert user_ids_found == set(THREADS_USER_IDS)


def test_threads_profile_posts(temp_dir):
    output_filename = temp_dir / "threads-profile-posts-1.csv"
    assert not output_filename.exists()
    max_posts = 50
    max_posts_per_user = 10
    collect_profile_posts(
        client_class=Threads,
        usernames_or_ids=THREADS_USER_MIX,
        output_csv=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        max_posts_per_user=max_posts_per_user,
        max_posts=max_posts,
        quiet=True,
    )
    assert output_filename.exists()
    with output_filename.open() as fobj:
        data = list(csv.DictReader(fobj))
    counter = Counter(row["user_id"] for row in data)
    user_ids_found = set(counter.keys())
    assert user_ids_found.issubset(THREADS_USER_IDS)
    assert len(data) == max_posts
    assert set(counter.values()) == {max_posts_per_user}


def test_threads_archive(temp_dir):
    output_filename = temp_dir / "threads-archive.zip"
    assert not output_filename.exists()
    max_posts = 20
    max_posts_per_user = 5
    archive(
        client_class=Threads,
        usernames_or_ids=THREADS_USER_MIX,
        output_zip=output_filename,
        min_wait=MIN_WAIT,
        max_wait=MAX_WAIT,
        max_posts_per_user=max_posts_per_user,
        max_posts=max_posts,
        quiet=True,
    )
    assert output_filename.exists()
    zf = ZipFile(output_filename)
    filenames = [fileinfo.filename for fileinfo in zf.filelist]
    assert "profile.csv" in filenames
    assert "post.csv" in filenames
    assert sum(1 for filename in filenames if filename.startswith("media/profile_")) > 0
    assert sum(1 for filename in filenames if filename.startswith("media/post_")) > 0



import random
import re
import time
from textwrap import dedent
from typing import Callable

REGEXP_LSD = re.compile('"l":"([^"]+)"}')


def random_waiter(min_wait: float, max_wait: float) -> Callable:

    def func(last_wait):
        this_wait = min_wait + random.random() * (max_wait - min_wait)
        while time.time() - last_wait < this_wait:
            time.sleep(this_wait / 10)

    return func


def default_browser_headers(user_agent: str) -> dict[str, str]:
    return {
        "accept": "*/*",
        "accept-language": "en-US;q=0.9,en;q=0.8,es-AR;q=0.7,es;q=0.6,it;q=0.5,pt-BR;q=0.4",
        "cache-control": "no-cache",
        "origin": "https://www.instagram.com",
        "pragma": "no-cache",
        "sec-ch-prefers-color-scheme": "light",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
        "user-agent": user_agent,
    }


def random_user_agent() -> str:
    """Return a random user agent from a curated real-world list of Linux, Windows and Mac desktop systems"""
    user_agents = (
        dedent(
            """
            Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36
            Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36
            Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.2 Safari/605.1.15
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36 Edg/108.0.1418.42'
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36 OPR/114.0.0.0
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36 OPR/115.0.0.0 (Edition std-2)
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 Edg/131.0.0.0
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 OPR/116.0.0.0
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36 Edg/132.0.0.0
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.130 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.69 Safari/537.36
            Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0
            Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:134.0) Gecko/20100101 Firefox/134.0
            Mozilla/5.0 (Windows NT 11.0; WOW64; rv:111.0) Gecko/20250101 Firefox/111.0
            Mozilla/5.0 (Windows NT 6.1; WOW64; rv:43.0) Gecko/20100101 Firefox/43.0
            Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:109.0) Gecko/20100101 Firefox/115.0
            Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US; rv:1.9.1.5) Gecko/20091102 Firefox/3.5.5 (.NET CLR 3.5.30729)
            Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36
            Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36 OPR/117.0.0.0
            Mozilla/5.0 (X11; Linux x86_64; rv:134.0) Gecko/20100101 Firefox/134.0
            Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:35.0) Gecko/20100101 Firefox/35.0
        """
        )
        .strip()
        .splitlines()
    )
    return random.choice(user_agents)
